package com.publicobject.issuesbrowser.swing;

import ca.odell.glazedlists.EventList;
import ca.odell.glazedlists.GroupingList;
import ca.odell.glazedlists.swing.GlazedListsSwing;
import ca.odell.glazedlists.event.ListEvent;
import ca.odell.glazedlists.event.ListEventListener;
import ca.odell.glazedlists.matchers.AbstractMatcherEditor;
import ca.odell.glazedlists.matchers.Matcher;
import com.publicobject.issuesbrowser.Issue;
import com.publicobject.issuesbrowser.IssueStatusComparator;
import javax.swing.*;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.text.MessageFormat;
import java.util.*;
import java.util.List;

/**
 * A MatcherEditor that produces Matchers that filter the issues based on the selected statuses.
 */
class StatusMatcherEditor extends AbstractMatcherEditor implements ListEventListener, ActionListener {
    
    /**
     * A MessageFormat to generate pretty names for our CheckBoxes which include the number of bugs with that status. 
     */
    private static final MessageFormat checkboxFormat = new MessageFormat("{0} {1,choice,0#|0<({1})}");
    
    /**
     * A panel housing a checkbox for each status. 
     */
    private JPanel checkBoxPanel = new JPanel(new GridLayout(2, 2));
    
    /**
     * A checkbox for each displayed status. 
     */
    private final Map statusCheckBoxes = new LinkedHashMap();
    
    /**
     * A user friendly String for each displayed status. 
     */
    private final Map statusLabels = new HashMap();
    
    /**
     * Issues grouped together by status. 
     */
    private final EventList issuesByStatus;
    
    /**
     * A cache of the list of statuses that mirrors the statuses of the issuesByStatus List.
     * It is used to determine which status is deleted when DELETE events arrive.
     */
    private List statuses = new ArrayList();
    
    public StatusMatcherEditor(EventList issues) {
        super();
        this.issuesByStatus = GlazedListsSwing.swingThreadProxyList(new GroupingList(issues, new IssueStatusComparator()));
        this.issuesByStatus.addListEventListener(this);
        this.statusCheckBoxes.put("NEW", buildCheckBox("New"));
        this.statusCheckBoxes.put("RESOLVED", buildCheckBox("Resolved"));
        this.statusCheckBoxes.put("STARTED", buildCheckBox("Started"));
        this.statusCheckBoxes.put("CLOSED", buildCheckBox("Closed"));
        this.statusLabels.put("NEW", "New");
        this.statusLabels.put("RESOLVED", "Resolved");
        this.statusLabels.put("STARTED", "Started");
        this.statusLabels.put("CLOSED", "Closed");
        this.checkBoxPanel.setOpaque(false);
        for (Iterator iter = statusCheckBoxes.values().iterator(); iter.hasNext(); ) {
            JCheckBox checkBox = (JCheckBox)iter.next();
            checkBox.addActionListener(this);
            this.checkBoxPanel.add(checkBox);
        }
    }
    
    /**
     * Returns the component responsible for editing the status filter.
     */
    public Component getComponent() {
        return this.checkBoxPanel;
    }
    
    /**
     * A convenience method to build a status checkbox with the given name.
     */
    private static JCheckBox buildCheckBox(String name) {
        final JCheckBox checkBox = new JCheckBox(name, true);
        checkBox.setOpaque(false);
        checkBox.setFocusable(false);
        checkBox.setMargin(new Insets(0, 0, 0, 0));
        return checkBox;
    }
    
    /**
     * Returns a StatusMatcher which matches Issues if their status is one
     * of the selected statuses.
     */
    private StatusMatcher buildMatcher() {
        final Set allowedStates = new HashSet();
        for (Iterator iter = statusCheckBoxes.entrySet().iterator(); iter.hasNext(); ) {
            Map.Entry entry = (Map.Entry)iter.next();
            if (((JCheckBox)entry.getValue()).isSelected()) allowedStates.add(entry.getKey());
        }
        return new StatusMatcher(allowedStates);
    }
    
    public void listChanged(ListEvent listChanges) {
        while (listChanges.next()) {
            final int type = listChanges.getType();
            final int index = listChanges.getIndex();
            final String status;
            final int count;
            if (type == ListEvent.INSERT) {
                List issuesOfThisStatus = (List)(List)issuesByStatus.get(index);
                status = ((Issue)(Issue)issuesOfThisStatus.get(0)).getStatus();
                statuses.add(index, status);
                count = issuesOfThisStatus.size();
            } else if (type == ListEvent.UPDATE) {
                List issuesOfThisStatus = (List)(List)issuesByStatus.get(index);
                status = (String)(String)statuses.get(index);
                count = issuesOfThisStatus.size();
            } else if (type == ListEvent.DELETE) {
                status = (String)statuses.remove(index);
                count = 0;
            } else {
                throw new IllegalStateException();
            }
            final JCheckBox checkBox = (JCheckBox)statusCheckBoxes.get(status);
            if (checkBox == null) continue;
            checkBox.setText(checkboxFormat.format(new Object[]{statusLabels.get(status), new Integer(count)}));
        }
    }
    
    public void actionPerformed(ActionEvent e) {
        final boolean isCheckBoxSelected = ((JCheckBox)(JCheckBox)e.getSource()).isSelected();
        final StatusMatcher statusMatcher = this.buildMatcher();
        if (statusMatcher.getStateCount() == 0) this.fireMatchNone(); else if (statusMatcher.getStateCount() == this.statusCheckBoxes.size()) this.fireMatchAll(); else if (isCheckBoxSelected) this.fireRelaxed(statusMatcher); else this.fireConstrained(statusMatcher);
    }
    
    /**
     * A StatusMatcher returns <tt>true</tt> if the status of the Issue is
     * one of the viewable status selected by the user.
     */
    private static class StatusMatcher implements Matcher {
        private final Set allowedStatuses;
        
        public StatusMatcher(Set allowedStatuses) {
            super();
            this.allowedStatuses = allowedStatuses;
        }
        
        public int getStateCount() {
            return this.allowedStatuses.size();
        }
        
        public boolean matches(Object x0) {
            Issue issue = (Issue)x0;
            return this.allowedStatuses.contains(issue.getStatus());
        }
        /*missing*/
    }
}
