package ca.odell.glazedlists;

import ca.odell.glazedlists.event.ListEvent;

/**
 * This {@link EventList} shows values from a continuous range of indices from
 * a source {@link EventList}. It can be used to limit the length of a list to
 * a desired size.
 *
 * <p><table border="1" width="100%" cellpadding="3" cellspacing="0">
 * <tr class="tableheadingcolor"><td colspan=2><font size="+2"><b>EventList Overview</b></font></td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Writable:</b></td><td>yes</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Concurrency:</b></td><td>thread ready, not thread safe</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Performance:</b></td><td>reads: O(1), writes O(1), change range O(1)</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Memory:</b></td><td>0 bytes per element</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Unit Tests:</b></td><td>N/A</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Issues:</b></td><td>
 *   <a href="https://glazedlists.dev.java.net/issues/show_bug.cgi?id=238">238</a>
 *   <a href="https://glazedlists.dev.java.net/issues/show_bug.cgi?id=278">278</a>
 * </td></tr>
 * </table>
 *
 * @author <a href="mailto:jesse@swank.ca">Jesse Wilson</a>
 */
public class RangeList extends TransformedList {
    
    /**
     * the user-specified range of the source list to include 
     */
    private int desiredStart = 0;
    private int desiredEnd = -1;
    
    /**
     * the first index in this list, inclusive 
     */
    private int currentStartIndex;
    
    /**
     * the last index in this list, exclusive 
     */
    private int currentEndIndex;
    
    /**
     * Create a new {@link RangeList} that limits the specified {@link EventList}
     * to a desired range.
     */
    public RangeList(EventList source) {
        super(source);
        currentStartIndex = 0;
        currentEndIndex = source.size();
        source.addListEventListener(this);
    }
    
    /**
     * {@inheritDoc} 
     */
    public final void listChanged(ListEvent listChanges) {
        updates.beginEvent(true);
        while (listChanges.next()) {
            int changeType = listChanges.getType();
            int changeIndex = listChanges.getIndex();
            if (changeType == ListEvent.DELETE) {
                if (changeIndex < currentStartIndex) {
                    currentStartIndex--;
                    currentEndIndex--;
                } else if (changeIndex < currentEndIndex) {
                    currentEndIndex--;
                    updates.addDelete(changeIndex - currentStartIndex);
                }
            } else if (changeType == ListEvent.INSERT) {
                if (changeIndex < currentStartIndex) {
                    currentStartIndex++;
                    currentEndIndex++;
                } else if (changeIndex < currentEndIndex) {
                    currentEndIndex++;
                    updates.addInsert(changeIndex - currentStartIndex);
                }
            } else if (changeType == ListEvent.UPDATE) {
                if (changeIndex >= currentStartIndex && changeIndex < currentEndIndex) {
                    updates.addUpdate(changeIndex - currentStartIndex);
                }
            }
        }
        adjustRange();
        updates.commitEvent();
    }
    
    /**
     * Set the range of values displayed by this {@link RangeList}.
     *
     * @param startIndex the first index of the source {@link EventList} to show, inclusive
     * @param endIndex the last index of the source {@link EventList} to show, exclusive
     */
    public void setRange(int startIndex, int endIndex) {
        this.desiredStart = startIndex;
        this.desiredEnd = endIndex;
        adjustRange();
    }
    
    /**
     * Set the range to include the specified indicies, offset from the end of
     * the source {@link EventList}. For example, to show the last five values, use:
     * <code>RangeList.setTailRange(5, 0);</code>
     *
     * <p>To include the 3rd last and 2nd last values, use:
     * <code>RangeList.setTailRange(3, 1);</code>.
     */
    public void setTailRange(int startIndex, int endIndex) {
        this.desiredStart = -startIndex - 1;
        this.desiredEnd = -endIndex - 1;
        adjustRange();
    }
    
    /**
     * Adjust the range of the {@link RangeList} in response to changes in the
     * source list or the desired start and end indices.
     */
    protected final void adjustRange() {
        updates.beginEvent(true);
        int desiredStartIndex = getStartIndex();
        int desiredEndIndex = getEndIndex();
        if (desiredEndIndex < desiredStartIndex) {
            int temp = desiredEndIndex;
            desiredEndIndex = desiredStartIndex;
            desiredStartIndex = temp;
        }
        if (desiredStartIndex < currentStartIndex) {
            updates.addInsert(0, currentStartIndex - desiredStartIndex - 1);
        } else if (currentStartIndex < desiredStartIndex && currentStartIndex < currentEndIndex) {
            int deleteThru = Math.min(desiredStartIndex, currentEndIndex);
            updates.addDelete(0, deleteThru - currentStartIndex - 1);
        }
        currentStartIndex = desiredStartIndex;
        if (desiredEndIndex < currentEndIndex) {
            int deleteFrom = Math.max(desiredEndIndex, currentStartIndex);
            updates.addDelete(deleteFrom - currentStartIndex, currentEndIndex - currentStartIndex - 1);
        } else if (currentEndIndex < desiredEndIndex && desiredStartIndex < desiredEndIndex) {
            int insertFrom = Math.max(currentEndIndex, currentStartIndex);
            updates.addInsert(insertFrom - currentStartIndex, desiredEndIndex - currentStartIndex - 1);
        }
        currentEndIndex = desiredEndIndex;
        updates.commitEvent();
    }
    
    /**
     * {@inheritDoc} 
     */
    public final int size() {
        return currentEndIndex - currentStartIndex;
    }
    
    /**
     * {@inheritDoc} 
     */
    protected final int getSourceIndex(int mutationIndex) {
        return mutationIndex + currentStartIndex;
    }
    
    /**
     * {@inheritDoc} 
     */
    protected final boolean isWritable() {
        return true;
    }
    
    /**
     * Get the first index of the source {@link EventList}
     * that is presented in this {@link RangeList}.
     */
    public int getStartIndex() {
        int desiredStartIndex = desiredStart >= 0 ? desiredStart : source.size() + desiredStart + 1;
        if (desiredStartIndex < 0) return 0;
        if (desiredStartIndex > source.size()) return source.size();
        return desiredStartIndex;
    }
    
    /**
     * Get the first index of the source {@link EventList}
     * that is beyond the range of this {@link RangeList}.
     */
    public int getEndIndex() {
        int desiredEndIndex = desiredEnd >= 0 ? desiredEnd : source.size() + desiredEnd + 1;
        int desiredStartIndex = getStartIndex();
        if (desiredEndIndex < desiredStartIndex) return desiredStartIndex;
        if (desiredEndIndex > source.size()) return source.size();
        return desiredEndIndex;
    }
}
