package ca.odell.glazedlists;

import ca.odell.glazedlists.event.ListEvent;
import ca.odell.glazedlists.impl.adt.IndexedTreeNode;
import ca.odell.glazedlists.impl.adt.IndexedTreeIterator;
import ca.odell.glazedlists.impl.adt.IndexedTree;
import ca.odell.glazedlists.impl.GlazedListsImpl;
import java.util.*;

/**
 * An {@link EventList} that shows its source {@link EventList} in sorted order.
 *
 * <p>The sorting strategy is specified with a {@link Comparator}. If no
 * {@link Comparator} is specified, all of the elements of the source {@link EventList}
 * must implement {@link Comparable}.
 *
 * <p>This {@link EventList} supports all write operations.
 *
 * <p><strong><font color="#FF0000">Warning:</font></strong> This class
 * breaks the contract required by {@link java.util.List}. See {@link EventList}
 * for an example.
 *
 * <p><table border="1" width="100%" cellpadding="3" cellspacing="0">
 * <tr class="tableheadingcolor"><td colspan=2><font size="+2"><b>EventList Overview</b></font></td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Writable:</b></td><td>yes</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Concurrency:</b></td><td>thread ready, not thread safe</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Performance:</b></td><td>reads: O(log N), writes O(log N), change comparator O(N log N)</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Memory:</b></td><td>72 bytes per element</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Unit Tests:</b></td><td>N/A</td></tr>
 * <tr><td class="tablesubheadingcolor"><b>Issues:</b></td><td>
 *   <a href="https://glazedlists.dev.java.net/issues/show_bug.cgi?id=39">39</a>
 *   <a href="https://glazedlists.dev.java.net/issues/show_bug.cgi?id=40">40</a>
 *   <a href="https://glazedlists.dev.java.net/issues/show_bug.cgi?id=58">58</a>
 *   <a href="https://glazedlists.dev.java.net/issues/show_bug.cgi?id=60">60</a>
 *   <a href="https://glazedlists.dev.java.net/issues/show_bug.cgi?id=62">62</a>
 *   <a href="https://glazedlists.dev.java.net/issues/show_bug.cgi?id=66">66</a>
 *   <a href="https://glazedlists.dev.java.net/issues/show_bug.cgi?id=161">161</a>
 *   <a href="https://glazedlists.dev.java.net/issues/show_bug.cgi?id=170">170</a>
 *   <a href="https://glazedlists.dev.java.net/issues/show_bug.cgi?id=206">206</a>
 *   <a href="https://glazedlists.dev.java.net/issues/show_bug.cgi?id=239">239</a>
 *   <a href="https://glazedlists.dev.java.net/issues/show_bug.cgi?id=255">255</a>
 *   <a href="https://glazedlists.dev.java.net/issues/show_bug.cgi?id=261">261</a>
 * </td></tr>
 * </table>
 *
 * @author <a href="mailto:jesse@odel.on.ca">Jesse Wilson</a>
 */
public final class SortedList extends TransformedList {
    
    /**
     * Sorting mode where elements are always in sorted order, even if this
     * requires that elements be moved from one index to another when their
     * value is changed.
     */
    public static final int STRICT_SORT_ORDER = 0;
    
    /**
     * Sorting mode where elements aren't moved when their value is changed,
     * even if this means they are no longer in perfect sorted order. This mode
     * is useful in editable lists and tables because it is annoying
     * for the current element to move if its value changes.
     */
    public static final int AVOID_MOVING_ELEMENTS = 1;
    
    /**
     * a map from the unsorted index to the sorted index 
     */
    private IndexedTree unsorted = null;
    
    /**
     * a map from the sorted index to the unsorted index 
     */
    private IndexedTree sorted = null;
    
    /**
     * the comparator that this list uses for sorting 
     */
    private Comparator comparator = null;
    
    /**
     * one of {@link STRICT_SORT_ORDER} or {@link AVOID_MOVING_ELEMENTS}. 
     */
    private int mode = STRICT_SORT_ORDER;
    
    /**
     * Creates a {@link SortedList} that sorts the specified {@link EventList}.
     * Because this constructor takes no {@link Comparator} argument, all
     * elements in the specified {@link EventList} must implement {@link Comparable}
     * or a {@link ClassCastException} will be thrown.
     */
    public SortedList(EventList source) {
        this(source, (Comparator)GlazedLists.comparableComparator());
    }
    
    /**
     * Creates a {@link SortedList} that sorts the specified {@link EventList}
     * using the specified {@link Comparator} to determine sort order. If the
     * specified {@link Comparator} is <code>null</code>, then this {@link List}
     * will be unsorted.
     */
    public SortedList(EventList source, Comparator comparator) {
        super(source);
        setComparator(comparator);
        source.addListEventListener(this);
    }
    
    /**
     * Modify the behaviour of this {@link SortedList} to one of the predefined modes.
     *
     * @param mode either {@link STRICT_SORT_ORDER} or {@link AVOID_MOVING_ELEMENTS}.
     */
    public void setMode(int mode) {
        if (mode != STRICT_SORT_ORDER && mode != AVOID_MOVING_ELEMENTS) throw new IllegalArgumentException("Mode must be either SortedList.STRICT_SORT_ORDER or SortedList.AVOID_MOVING_ELEMENTS");
        if (mode == this.mode) return;
        this.mode = mode;
        if (this.mode == STRICT_SORT_ORDER) {
            setComparator(getComparator());
        }
    }
    
    /**
     * Get the behaviour mode for this {@link SortedList}.
     *
     * @return one of {@link STRICT_SORT_ORDER} (default) or
     *     {@link AVOID_MOVING_ELEMENTS}.
     */
    public int getMode() {
        return this.mode;
    }
    
    /**
     * {@inheritDoc} 
     */
    public void listChanged(ListEvent listChanges) {
        if (listChanges.isReordering()) {
            int[] reorderMap = listChanges.getReorderMap();
            IndexedTreeNode[] sortedNodes = new IndexedTreeNode[sorted.size()];
            int index = 0;
            for (IndexedTreeIterator i = unsorted.iterator(0); i.hasNext(); index++) {
                IndexedTreeNode unsortedNode = i.next();
                sortedNodes[index] = (IndexedTreeNode)unsortedNode.getValue();
            }
            index = 0;
            for (IndexedTreeIterator i = unsorted.iterator(0); i.hasNext(); index++) {
                IndexedTreeNode unsortedNode = i.next();
                unsortedNode.setValue(sortedNodes[reorderMap[index]]);
                sortedNodes[reorderMap[index]].setValue(unsortedNode);
            }
            return;
        }
        updates.beginEvent();
        LinkedList insertNodes = new LinkedList();
        List updateNodes = new ArrayList();
        while (listChanges.next()) {
            int unsortedIndex = listChanges.getIndex();
            int changeType = listChanges.getType();
            if (changeType == ListEvent.INSERT) {
                IndexedTreeNode unsortedNode = unsorted.addByNode(unsortedIndex, IndexedTreeNode.EMPTY_NODE);
                insertNodes.addLast(unsortedNode);
            } else if (changeType == ListEvent.UPDATE) {
                IndexedTreeNode unsortedNode = unsorted.getNode(unsortedIndex);
                IndexedTreeNode sortedNode = (IndexedTreeNode)unsortedNode.getValue();
                sortedNode.setSorted(false);
                updateNodes.add(sortedNode);
            } else if (changeType == ListEvent.DELETE) {
                IndexedTreeNode unsortedNode = unsorted.getNode(unsortedIndex);
                unsortedNode.removeFromTree(unsorted);
                int deleteSortedIndex = deleteByUnsortedNode(unsortedNode);
                updates.addDelete(deleteSortedIndex);
            }
        }
        for (Iterator i = updateNodes.iterator(); i.hasNext(); ) {
            IndexedTreeNode sortedNode = (IndexedTreeNode)i.next();
            int originalIndex = sortedNode.getIndex();
            if (isNodeInSortedOrder(sortedNode)) {
                sortedNode.setSorted(true);
                updates.addUpdate(originalIndex);
            } else if (mode == AVOID_MOVING_ELEMENTS) {
                updates.addUpdate(originalIndex);
            } else {
                sortedNode.removeFromTree(sorted);
                updates.addDelete(originalIndex);
                int insertedIndex = insertByUnsortedNode((IndexedTreeNode)sortedNode.getValue());
                updates.addInsert(insertedIndex);
            }
        }
        while (!insertNodes.isEmpty()) {
            IndexedTreeNode insertNode = (IndexedTreeNode)insertNodes.removeFirst();
            int insertedIndex = insertByUnsortedNode(insertNode);
            updates.addInsert(insertedIndex);
        }
        updates.commitEvent();
    }
    
    /**
     * Whether this node is greater or equal to its neighbour on the left and
     * less than or equal to its neighbour on the right.
     *
     * <p>This method skips unsorted nodes, whose value should not be compared
     * against when determining tree ordering.
     */
    private boolean isNodeInSortedOrder(IndexedTreeNode sortedNode) {
        for (IndexedTreeNode leftNeighbour = sortedNode.previous(); leftNeighbour != null; leftNeighbour = leftNeighbour.previous()) {
            if (!leftNeighbour.isSorted()) continue;
            if (sorted.getComparator().compare(leftNeighbour.getValue(), sortedNode.getValue()) > 0) return false;
            break;
        }
        for (IndexedTreeNode rightNeighbour = sortedNode.next(); rightNeighbour != null; rightNeighbour = rightNeighbour.next()) {
            if (!rightNeighbour.isSorted()) continue;
            if (sorted.getComparator().compare(sortedNode.getValue(), rightNeighbour.getValue()) > 0) return false;
            break;
        }
        return true;
    }
    
    /**
     * Inserts the specified unsorted node as the value in the sorted tree
     * and returns the sorted order.
     *
     * @return the sortIndex of the inserted object.
     */
    private int insertByUnsortedNode(IndexedTreeNode unsortedNode) {
        IndexedTreeNode sortedNode = sorted.addByNode(unsortedNode);
        unsortedNode.setValue(sortedNode);
        return sortedNode.getIndex();
    }
    
    /**
     * Deletes the node in the sorted tree based on the value of the specified
     * unsorted tree node.
     *
     * @return the sortIndex of the deleted object.
     */
    private int deleteByUnsortedNode(IndexedTreeNode unsortedNode) {
        IndexedTreeNode sortedNode = (IndexedTreeNode)(IndexedTreeNode)unsortedNode.getValue();
        int sortedIndex = sortedNode.getIndex();
        sortedNode.removeFromTree(sorted);
        return sortedIndex;
    }
    
    /**
     * {@inheritDoc} 
     */
    protected int getSourceIndex(int mutationIndex) {
        IndexedTreeNode sortedNode = sorted.getNode(mutationIndex);
        IndexedTreeNode unsortedNode = (IndexedTreeNode)(IndexedTreeNode)sortedNode.getValue();
        return unsortedNode.getIndex();
    }
    
    /**
     * {@inheritDoc} 
     */
    protected boolean isWritable() {
        return true;
    }
    
    /**
     * Gets the {@link Comparator} that is being used to sort this list.
     *
     * @return the {@link Comparator} in use, or <tt>null</tt> if this list is
     *      currently unsorted. If this is an {@link ca.odell.glazedlists.EventList} of {@link Comparable}
     *      elements in natural order, then a {@link ca.odell.glazedlists.impl.sort.ComparableComparator} will
     *      be returned.
     */
    public Comparator getComparator() {
        return comparator;
    }
    
    /**
     * Set the {@link Comparator} in use in this {@link ca.odell.glazedlists.EventList}. This will
     * sort the {@link ca.odell.glazedlists.EventList} into a new order.
     *
     * <p>Performance Note: sorting will take <code>O(N * Log N)</code> time.
     *
     * <p><strong><font color="#FF0000">Warning:</font></strong> This method is
     * thread ready but not thread safe. See {@link ca.odell.glazedlists.EventList} for an example
     * of thread safe code.
     *
     * @param comparator the {@link Comparator} to specify how to sort the list. If
     *      the source {@link ca.odell.glazedlists.EventList} elements implement {@link Comparable},
     *      you may use a {@link ca.odell.glazedlists.impl.sort.ComparableComparator} to sort them in their
     *      natural order. You may also specify <code>null</code> to put this
     *      {@link SortedList} in unsorted order.
     */
    public void setComparator(Comparator comparator) {
        this.comparator = comparator;
        IndexedTree previousSorted = sorted;
        Comparator treeComparator = null;
        if (comparator != null) treeComparator = new IndexedTreeNodeComparator(comparator); else treeComparator = new IndexedTreeNodeRawOrderComparator();
        sorted = new IndexedTree(treeComparator);
        if (previousSorted == null && unsorted == null) {
            unsorted = new IndexedTree();
            for (int i = 0; i < source.size(); i++) {
                IndexedTreeNode unsortedNode = unsorted.addByNode(i, IndexedTreeNode.EMPTY_NODE);
                insertByUnsortedNode(unsortedNode);
            }
            return;
        }
        if (source.size() == 0) return;
        for (IndexedTreeIterator i = unsorted.iterator(0); i.hasNext(); ) {
            IndexedTreeNode unsortedNode = i.next();
            insertByUnsortedNode(unsortedNode);
        }
        int[] reorderMap = new int[size()];
        int oldSortedIndex = 0;
        for (IndexedTreeIterator i = previousSorted.iterator(0); i.hasNext(); oldSortedIndex++) {
            IndexedTreeNode oldSortedNode = i.next();
            IndexedTreeNode unsortedNode = (IndexedTreeNode)(IndexedTreeNode)oldSortedNode.getValue();
            IndexedTreeNode newSortedNode = (IndexedTreeNode)(IndexedTreeNode)unsortedNode.getValue();
            int newSortedIndex = newSortedNode.getIndex();
            reorderMap[newSortedIndex] = oldSortedIndex;
        }
        updates.beginEvent();
        updates.reorder(reorderMap);
        updates.commitEvent();
    }
    
    /**
     * {@inheritDoc} 
     */
    public int indexOf(Object object) {
        if (mode != STRICT_SORT_ORDER || comparator == null) return source.indexOf(object);
        int index = sorted.indexOf(object);
        if (index == -1) return -1;
        for (; index < size(); index++) {
            Object objectAtIndex = get(index);
            if (comparator.compare((Object)object, objectAtIndex) != 0) return -1;
            if (GlazedListsImpl.equal(object, objectAtIndex)) return index;
        }
        return -1;
    }
    
    /**
     * {@inheritDoc} 
     */
    public int lastIndexOf(Object object) {
        if (mode != STRICT_SORT_ORDER || comparator == null) return source.lastIndexOf(object);
        int index = sorted.lastIndexOf(object);
        if (index == -1) return -1;
        for (; index > -1; index--) {
            Object objectAtIndex = get(index);
            if (comparator.compare((Object)object, objectAtIndex) != 0) return -1;
            if (GlazedListsImpl.equal(object, objectAtIndex)) return index;
        }
        return -1;
    }
    
    /**
     * Returns the first index of the <code>object</code>'s sort location or
     * the first index at which the <code>object</code> could be positioned if
     * inserted.
     *
     * <p>Unlike {@link #indexOf} this method does not guarantee the given
     * <code>object</code> {@link Object#equals(Object) equals} the element at
     * the returned index. Instead, they are indistinguishable according to the
     * sorting {@link Comparator}.
     *
     * @return a value in <tt>[0, size()]</tt> inclusive
     */
    public int sortIndex(Object object) {
        if (comparator == null) throw new IllegalStateException("No Comparator exists to perform this operation");
        final int sortIndex = sorted.indexOf(object);
        return sortIndex == -1 ? sorted.indexOfSimulated(object) : sortIndex;
    }
    
    /**
     * Returns the last index of the <code>object</code>'s sort location or
     * the last index at which the <code>object</code> could be positioned if
     * inserted.
     *
     * <p>Unlike {@link #lastIndexOf} this method does not guarantee the given
     * <code>object</code> {@link Object#equals(Object) equals} the element at
     * the returned index. Instead, they are indistinguishable according to the
     * sorting {@link Comparator}.
     *
     * @return a value in <tt>[0, size()]</tt> inclusive
     */
    public int lastSortIndex(Object object) {
        if (comparator == null) throw new IllegalStateException("No Comparator exists to perform this operation");
        final int lastSortIndex = sorted.lastIndexOf(object);
        return lastSortIndex == -1 ? sorted.indexOfSimulated(object) : lastSortIndex;
    }
    
    /**
     * Returns the index in this list of the first occurrence of the specified
     * element, or the index where that element would be in the list if it were
     * inserted.
     *
     * @return the index in this list of the first occurrence of the specified
     *      element, or the index where that element would be in the list if it
     *      were inserted. This will return a value in <tt>[0, size()]</tt>,
     *      inclusive.
     *
     * @deprecated Deprecated as of 12/11/2005. Replaced with {@link #sortIndex(Object)}
     *      which has cleaner semantics.
     */
    public int indexOfSimulated(Object object) {
        return comparator != null ? sorted.indexOfSimulated(object) : size();
    }
    
    /**
     * {@inheritDoc} 
     */
    public boolean contains(Object object) {
        return indexOf(object) != -1;
    }
    
    /**
     * A comparator that takes an indexed node, and compares the value
     * of an object in a list that has the index of that node.
     *
     * <p>If one of the objects passed to {@link #compare} is not an
     * {@link IndexedTreeNode}, it will compare the object directly to the object
     * in the source {@link ca.odell.glazedlists.EventList} referenced by the {@link IndexedTreeNode}.
     * This functionality is necessary to allow use of the underlying
     * {@link Comparator} within {@link IndexedTree} to support {@link java.util.List#indexOf},
     * {@link java.util.List#lastIndexOf}, and {@link java.util.List#contains}.
     */
    private class IndexedTreeNodeComparator implements Comparator {
        
        /**
         * the actual comparator used on the values found 
         */
        private Comparator comparator;
        
        /**
         * Creates an {@link IndexedTreeNodeComparator} that compares the
         * objects in the source list based on the indexes of the tree
         * nodes being compared.
         */
        public IndexedTreeNodeComparator(Comparator comparator) {
            super();
            this.comparator = comparator;
        }
        
        /**
         * Compares object alpha to object beta by using the source comparator.
         */
        public int compare(Object alpha, Object beta) {
            Object alphaObject = alpha;
            Object betaObject = beta;
            int alphaIndex = -1;
            int betaIndex = -1;
            if (alpha instanceof IndexedTreeNode) {
                IndexedTreeNode alphaTreeNode = (IndexedTreeNode)(IndexedTreeNode)alpha;
                alphaIndex = alphaTreeNode.getIndex();
                alphaObject = source.get(alphaIndex);
            }
            if (beta instanceof IndexedTreeNode) {
                IndexedTreeNode betaTreeNode = (IndexedTreeNode)(IndexedTreeNode)beta;
                betaIndex = betaTreeNode.getIndex();
                betaObject = source.get(betaIndex);
            }
            int result = comparator.compare(alphaObject, betaObject);
            if (result != 0) return result;
            if (alphaIndex != -1 && betaIndex != -1) return alphaIndex - betaIndex;
            return 0;
        }
    }
    
    /**
     * A comparator that takes an indexed node, and compares the index of that node.
     */
    private static class IndexedTreeNodeRawOrderComparator implements Comparator {
        
        private IndexedTreeNodeRawOrderComparator() {
            super();
        }
        
        /**
         * Compares the alpha object to the beta object by their indicies.
         */
        public int compare(Object alpha, Object beta) {
            try {
                IndexedTreeNode alphaTreeNode = (IndexedTreeNode)(IndexedTreeNode)alpha;
                IndexedTreeNode betaTreeNode = (IndexedTreeNode)(IndexedTreeNode)beta;
                int alphaIndex = alphaTreeNode.getIndex();
                int betaIndex = betaTreeNode.getIndex();
                return alphaIndex - betaIndex;
            } catch (ClassCastException e) {
                System.out.println(alpha.getClass());
                System.out.println(beta.getClass());
                throw e;
            }
        }
    }
    
    /**
     * {@inheritDoc} 
     */
    public Iterator iterator() {
        return new SortedListIterator();
    }
    
    /**
     * The fast iterator for SortedList
     */
    private class SortedListIterator implements Iterator {
        
        private SortedListIterator() {
            super();
        }
        
        /**
         * the IndexedTreeIterator to use to move across the tree 
         */
        private IndexedTreeIterator treeIterator = sorted.iterator(0);
        
        /**
         * Returns true iff there are more value to iterate on by caling next()
         */
        public boolean hasNext() {
            return treeIterator.hasNext();
        }
        
        /**
         * Returns the next value in the iteration.
         */
        public Object next() {
            IndexedTreeNode sortedNode = (IndexedTreeNode)treeIterator.next();
            IndexedTreeNode unsortedNode = (IndexedTreeNode)(IndexedTreeNode)sortedNode.getValue();
            return source.get(unsortedNode.getIndex());
        }
        
        /**
         * Removes the last value returned by this iterator.
         */
        public void remove() {
            int indexToRemove = treeIterator.previousIndex();
            SortedList.this.source.remove(getSourceIndex(indexToRemove));
            treeIterator = sorted.iterator(indexToRemove);
        }
    }
}
