package ca.odell.glazedlists.impl.adt;

import java.util.*;

/**
 * A BarcodeIterator is a specialized {@link Iterator} implementation for moving
 * over a Barcode efficiently.
 *
 * @author <a href="mailto:kevin@swank.ca">Kevin Maltby</a>
 */
public class BarcodeIterator implements Iterator {
    
    /**
     * keep a reference for removes in the trailing whitespace 
     */
    private Barcode barcode = null;
    
    /**
     * the current node being inspected 
     */
    private BarcodeNode currentNode = null;
    
    /**
     * the number of requests on the current node 
     */
    private int localIndex = -1;
    
    /**
     * the number of black elements before this node 
     */
    private int blackSoFar = 0;
    
    /**
     * the number of white elements before this node 
     */
    private int whiteSoFar = 0;
    
    /**
     * Creates a new Iterator for the given Barcode.
     */
    BarcodeIterator(Barcode barcode) {
        super();
        BarcodeNode root = barcode.getRootNode();
        if (root != null) {
            currentNode = root;
            while (currentNode.left != null) {
                currentNode = currentNode.left;
            }
        }
        this.barcode = barcode;
    }
    
    /**
     * Returns whether or not there are more values in the {@link Barcode} to
     * iterate over.
     */
    public boolean hasNext() {
        if (getIndex() == barcode.size() - 1) {
            return false;
        }
        return true;
    }
    
    /**
     * Returns true if there are more BLACK elements in the {@link Barcode} to
     * move the {@link Iterator} to.
     */
    public boolean hasNextBlack() {
        if (getIndex() >= barcode.treeSize() - 1) return false; else if (currentNode == null) return false; else return true;
    }
    
    /**
     * Returns true if there are more WHITE elements in the {@link Barcode} to
     * move the {@link Iterator} to.
     */
    public boolean hasNextWhite() {
        if (barcode.size() != barcode.treeSize()) return hasNext(); else if (currentNode == null) return false; else if (localIndex < currentNode.whiteSpace - 1 || whiteSoFar + currentNode.whiteSpace < barcode.whiteSize()) return true; else return false;
    }
    
    /**
     * Returns true if there are more elements in the {@link Barcode} to
     * move the {@link Iterator} to that match the provided colour.
     */
    public boolean hasNextColour(Object colour) {
        if (colour == Barcode.BLACK) return hasNextBlack();
        return hasNextWhite();
    }
    
    /**
     * Gets the next value in this SparseList.
     */
    public Object next() {
        localIndex++;
        if (currentNode == null) {
            if (getIndex() < barcode.size()) {
                return Barcode.WHITE;
            } else {
                throw new NoSuchElementException();
            }
        } else if (localIndex >= currentNode.whiteSpace + currentNode.rootSize) {
            if (getIndex() < barcode.treeSize()) {
                blackSoFar += currentNode.rootSize;
                whiteSoFar += currentNode.whiteSpace;
                findNextNode();
                localIndex = 0;
            } else {
                if (getIndex() < barcode.size()) {
                    return Barcode.WHITE;
                } else {
                    throw new NoSuchElementException();
                }
            }
        }
        if (localIndex < currentNode.whiteSpace) {
            return Barcode.WHITE;
        } else {
            return Barcode.BLACK;
        }
    }
    
    /**
     * Moves this {@link Iterator} to the next element in the {@link Barcode}
     * that is BLACK.
     *
     * @throws NoSuchElementException if hasNextBlack() returns false.
     */
    public Object nextBlack() {
        localIndex++;
        if (currentNode == null) {
            throw new NoSuchElementException();
        } else if (localIndex < currentNode.whiteSpace) {
            localIndex = currentNode.whiteSpace;
        } else if (localIndex >= currentNode.whiteSpace + currentNode.rootSize) {
            if (getIndex() < barcode.treeSize()) {
                whiteSoFar += currentNode.whiteSpace;
                blackSoFar += currentNode.rootSize;
                findNextNode();
                localIndex = currentNode.whiteSpace;
            } else {
                throw new NoSuchElementException();
            }
        }
        if (localIndex < currentNode.whiteSpace) throw new IllegalStateException();
        return Barcode.BLACK;
    }
    
    /**
     * Moves this {@link Iterator} to the next element in the {@link Barcode}
     * that is WHITE.
     *
     * @throws NoSuchElementException if hasNextWhite() returns false.
     */
    public Object nextWhite() {
        localIndex++;
        if (currentNode == null) {
            if (getIndex() < barcode.size()) {
                return Barcode.WHITE;
            } else {
                throw new NoSuchElementException();
            }
        } else if (localIndex >= currentNode.whiteSpace) {
            if (getIndex() < barcode.treeSize() && getIndex() + currentNode.rootSize >= barcode.treeSize()) {
                localIndex = currentNode.whiteSpace;
                localIndex += currentNode.rootSize;
            }
            if (getIndex() < barcode.treeSize()) {
                blackSoFar += currentNode.rootSize;
                whiteSoFar += currentNode.whiteSpace;
                findNextNode();
                localIndex = 0;
            } else {
                if (getIndex() < barcode.size()) {
                    return Barcode.WHITE;
                } else {
                    throw new NoSuchElementException();
                }
            }
        }
        if (localIndex >= currentNode.whiteSpace) throw new IllegalStateException();
        return Barcode.WHITE;
    }
    
    /**
     * Moves this {@link Iterator} to the next element in the {@link Barcode}
     * that matches the provided colour.
     *
     * @throws NoSuchElementException if hasNextColour(colour) returns false.
     */
    public Object nextColour(Object colour) {
        if (colour == Barcode.BLACK) return nextBlack();
        return nextWhite();
    }
    
    /**
     * Removes the current value at the Iterator from the {@link Barcode}.
     */
    public void remove() {
        if (localIndex == -1) {
            throw new NoSuchElementException("Cannot call remove() before next() is called.");
        } else if (currentNode == null || getIndex() >= barcode.treeSize()) {
            barcode.remove(getIndex(), 1);
            localIndex--;
        } else {
            BarcodeNode affectedNode = currentNode;
            if (localIndex == 0 && currentNode.whiteSpace == 1 && getIndex() != 0) {
                findPreviousNode();
                blackSoFar -= currentNode.rootSize;
                whiteSoFar -= currentNode.whiteSpace;
                localIndex += currentNode.whiteSpace + currentNode.rootSize;
            } else if (localIndex == currentNode.whiteSpace && currentNode.rootSize == 1) {
                if (localIndex == barcode.treeSize() - 1) {
                    currentNode = null;
                } else if (getIndex() == barcode.treeSize() - 1) {
                    findPreviousNode();
                    blackSoFar -= currentNode.rootSize;
                    whiteSoFar -= currentNode.whiteSpace;
                    localIndex += currentNode.whiteSpace + currentNode.rootSize;
                } else {
                    findNextNode();
                }
            }
            affectedNode.removeBaseCase(getIndex(), localIndex);
            localIndex--;
        }
    }
    
    /**
     * Sets the most recently viewed element to WHITE and returns the white-centric
     * index of the element after the set is complete.
     */
    public int setWhite() {
        if (localIndex == -1) {
            throw new NoSuchElementException("Cannot call setWhite() before next() is called.");
        } else if (currentNode == null || getIndex() >= barcode.treeSize() || localIndex < currentNode.whiteSpace) {
            return getWhiteIndex();
        } else if (getIndex() != barcode.treeSize() - 1) {
            if (currentNode.rootSize == 1) {
                BarcodeNode affectedNode = currentNode;
                findNextNode();
                affectedNode.setWhite(getIndex(), localIndex, 1);
                if (barcode.getRootNode() != null) barcode.treeSizeChanged();
                if (currentNode.whiteSpace == 0 && currentNode.rootSize == 0) currentNode = affectedNode;
                return whiteSoFar + localIndex;
            } else if (localIndex == currentNode.whiteSpace) {
                currentNode.setWhite(getIndex(), localIndex, 1);
                if (barcode.getRootNode() != null) barcode.treeSizeChanged();
                return whiteSoFar + localIndex;
            } else {
                currentNode.setWhite(getIndex(), localIndex, 1);
                if (barcode.getRootNode() != null) barcode.treeSizeChanged();
                blackSoFar += currentNode.rootSize;
                whiteSoFar += currentNode.whiteSpace;
                findNextNode();
                localIndex = 0;
                return whiteSoFar;
            }
        } else if (currentNode.rootSize == 1) {
            BarcodeNode affectedNode = currentNode;
            if (currentNode.whiteSpace + 1 == barcode.treeSize()) {
                currentNode = null;
                affectedNode.setWhite(getIndex(), localIndex, 1);
                if (barcode.getRootNode() != null) barcode.treeSizeChanged();
                return localIndex;
            } else {
                findPreviousNode();
                int currentLocalIndex = localIndex;
                blackSoFar -= currentNode.rootSize;
                whiteSoFar -= currentNode.whiteSpace;
                localIndex += currentNode.whiteSpace + currentNode.rootSize;
                affectedNode.setWhite(getIndex(), currentLocalIndex, 1);
                if (barcode.getRootNode() != null) barcode.treeSizeChanged();
                if (currentNode.whiteSpace == 0 && currentNode.rootSize == 0) currentNode = affectedNode;
                return whiteSoFar + localIndex - currentNode.rootSize;
            }
        } else {
            currentNode.setWhite(getIndex(), localIndex, 1);
            if (barcode.getRootNode() != null) barcode.treeSizeChanged();
            return whiteSoFar + localIndex - currentNode.rootSize;
        }
    }
    
    /**
     * Sets the most recently viewed element to BLACK and returns the white-centric
     * index of the element after the set is complete.
     */
    public int setBlack() {
        if (localIndex == -1) {
            throw new NoSuchElementException("Cannot call setBlack() before next() is called.");
        } else if (currentNode == null) {
            barcode.setBlack(getIndex(), 1);
            currentNode = barcode.getRootNode();
            return 0;
        } else if (getIndex() == barcode.treeSize()) {
            barcode.setBlack(getIndex(), 1);
            if (barcode.getRootNode() != null) barcode.treeSizeChanged();
        } else if (getIndex() > barcode.treeSize()) {
            barcode.setBlack(getIndex(), 1);
            if (barcode.getRootNode() != null) barcode.treeSizeChanged();
            whiteSoFar += currentNode.whiteSpace;
            blackSoFar += currentNode.rootSize;
            localIndex -= currentNode.whiteSpace + currentNode.rootSize;
            findNextNode();
        } else if (localIndex < currentNode.whiteSpace && currentNode.whiteSpace == 1) {
            if (getIndex() == 0) {
                currentNode.setBlack(getIndex(), localIndex, 1);
                if (barcode.getRootNode() != null) barcode.treeSizeChanged();
            } else {
                BarcodeNode affectedNode = currentNode;
                findPreviousNode();
                int currentLocalIndex = localIndex;
                blackSoFar -= currentNode.rootSize;
                whiteSoFar -= currentNode.whiteSpace;
                localIndex += currentNode.whiteSpace + currentNode.rootSize;
                affectedNode.setBlack(getIndex(), currentLocalIndex, 1);
                if (barcode.getRootNode() != null) barcode.treeSizeChanged();
                if (currentNode.whiteSpace == 0 && currentNode.rootSize == 0) currentNode = affectedNode;
            }
        } else if (localIndex < currentNode.whiteSpace) {
            if (localIndex == 0) {
                currentNode.setBlack(getIndex(), localIndex, 1);
                if (barcode.getRootNode() != null) barcode.treeSizeChanged();
                blackSoFar++;
                localIndex--;
                return blackSoFar - 1;
            } else if (localIndex == currentNode.whiteSpace - 1) {
                currentNode.setBlack(getIndex(), localIndex, 1);
                if (barcode.getRootNode() != null) barcode.treeSizeChanged();
                return blackSoFar;
            } else {
                currentNode.setBlack(getIndex(), localIndex, 1);
                if (barcode.getRootNode() != null) barcode.treeSizeChanged();
                whiteSoFar += localIndex;
                blackSoFar++;
                localIndex = -1;
                return blackSoFar - 1;
            }
        }
        return getBlackIndex();
    }
    
    /**
     * Sets the most recently viewed element to the value of colour and returns
     * the colour specific index of the element after the set is complete.
     */
    public int set(Object colour) {
        if (colour == Barcode.BLACK) return setBlack();
        return setWhite();
    }
    
    /**
     * Gets the index of the last element visited.
     */
    public int getIndex() {
        return blackSoFar + whiteSoFar + localIndex;
    }
    
    /**
     * Gets the black-centric index of the last element visited or -1 if that
     * element is white.
     */
    public int getBlackIndex() {
        if (localIndex == -1) {
            return blackSoFar - 1;
        } else if (currentNode == null || localIndex < currentNode.whiteSpace || localIndex >= currentNode.whiteSpace + currentNode.rootSize) {
            return -1;
        }
        return blackSoFar + localIndex - currentNode.whiteSpace;
    }
    
    /**
     * Gets the white-centric index of the last element visited or -1 if that
     * element is black.
     */
    public int getWhiteIndex() {
        if (currentNode == null) {
            if (localIndex == -1 && whiteSoFar != 0) return whiteSoFar - 1; else return localIndex;
        } else if (localIndex >= currentNode.whiteSpace && localIndex < currentNode.whiteSpace + currentNode.rootSize) return -1; else if (localIndex >= currentNode.whiteSpace + currentNode.rootSize) return whiteSoFar + localIndex - currentNode.rootSize;
        return whiteSoFar + localIndex;
    }
    
    /**
     * Gets the colour-centric index of the last element based on the value of
     * colour.  If the element didn't match the colour specified, this method
     * returns -1.
     */
    public int getColourIndex(Object colour) {
        if (colour == Barcode.WHITE) return getWhiteIndex();
        return getBlackIndex();
    }
    
    /**
     * Finds the next node in the tree.
     */
    private void findNextNode() {
        if (currentNode.right != null) {
            currentNode = currentNode.right;
            while (currentNode.left != null) {
                currentNode = currentNode.left;
            }
        } else if (currentNode.parent.left == currentNode) {
            currentNode = currentNode.parent;
        } else if (currentNode.parent.right == currentNode) {
            while (currentNode.parent.right == currentNode) {
                currentNode = currentNode.parent;
            }
            currentNode = currentNode.parent;
        } else {
            throw new IllegalStateException();
        }
    }
    
    /**
     * Finds the previous node in the tree.
     */
    private void findPreviousNode() {
        if (currentNode.left != null) {
            currentNode = currentNode.left;
            while (currentNode.right != null) {
                currentNode = currentNode.right;
            }
        } else if (currentNode.parent.right == currentNode) {
            currentNode = currentNode.parent;
        } else if (currentNode.parent.left == currentNode) {
            while (currentNode.parent.left == currentNode) {
                currentNode = currentNode.parent;
            }
            currentNode = currentNode.parent;
        } else {
            throw new IllegalStateException();
        }
    }
}
