package ca.odell.glazedlists.impl.ctp;

import ca.odell.glazedlists.impl.nio.*;
import java.util.*;
import java.nio.*;
import java.nio.channels.*;
import java.net.*;
import java.io.*;
import java.util.logging.*;

/**
 * The CTPConnectionManager provides managed access to multiple CTP connections
 * for both incoming and outgoing data.
 *
 * <p>Each instance of this class owns a single thread which is used to perform
 * all read and write operations on all connections. A pool of other threads are
 * used to notify the handlers of the data and status of a connection.
 *
 * @author <a href="mailto:jesse@odel.on.ca">Jesse Wilson</a>
 */
public final class CTPConnectionManager implements NIOServer {
    
    /**
     * default port to bind to 
     */
    private static final int DEFAULT_PORT = 5309;
    
    /**
     * port to listen for incoming connections 
     */
    private int listenPort = -1;
    
    /**
     * factory for handlers of incoming connections 
     */
    private CTPHandlerFactory handlerFactory;
    
    /**
     * the I/O event queue daemon 
     */
    private NIODaemon nioDaemon = null;
    
    /**
     * Creates a connection manager that handles incoming connections using the
     * specified connect handler. This binds to the default port.
     */
    public CTPConnectionManager(CTPHandlerFactory handlerFactory) {
        this(handlerFactory, DEFAULT_PORT);
    }
    
    /**
     * Creates a connection manager that handles incoming connections using the
     * specified connect handler. This binds to the specified port.
     */
    public CTPConnectionManager(CTPHandlerFactory handlerFactory, int listenPort) {
        super();
        this.handlerFactory = handlerFactory;
        this.listenPort = listenPort;
    }
    
    /**
     * Starts the CTPConnectionManager listening to incoming connections and
     * managing outgoing connections.
     *
     * @return true if the server successfully binds to the listen port.
     */
    public synchronized boolean start() throws IOException {
        if (nioDaemon != null) throw new IllegalStateException();
        nioDaemon = new NIODaemon();
        nioDaemon.start();
        try {
            nioDaemon.invokeAndWait(new StartServer(this, listenPort));
        } catch (RuntimeException e) {
            nioDaemon.stop();
            if (e.getCause() instanceof IOException) throw (IOException)(IOException)e.getCause(); else throw e;
        }
        return true;
    }
    
    /**
     * Stops the CTPConnectionManager and closes all connections.
     */
    public void stop() {
        nioDaemon.stop();
    }
    
    /**
     * Get the daemon that does all the threading and selection.
     */
    public NIODaemon getNIODaemon() {
        return nioDaemon;
    }
    
    /**
     * Handle an incoming connection.
     *
     * <p>This creates a CTPServerProtocol to handle the connection.
     *
     * @return the SelectionKey that is attached to the created connection.
     */
    public void handleAccept(SelectionKey key, Selector selector) {
        SocketChannel channel = null;
        SelectionKey channelKey = null;
        try {
            ServerSocketChannel server = (ServerSocketChannel)(ServerSocketChannel)key.channel();
            channel = server.accept();
            if (channel == null) return;
            channel.configureBlocking(false);
            channelKey = channel.register(selector, 0);
        } catch (IOException e) {
            return;
        }
        CTPHandler handler = handlerFactory.constructHandler();
        CTPConnection server = CTPConnection.server(channelKey, handler, this);
        channelKey.attach(server);
        server.handleConnect();
    }
    
    /**
     * Connect to the specified host.
     */
    public void connect(CTPHandler handler, String host, int port) {
        nioDaemon.invokeLater(new OpenConnection(this, handler, host, port));
    }
    
    public void connect(CTPHandler handler, String host) {
        connect(handler, host, DEFAULT_PORT);
    }
}
